/**
@file	Pileup.cpp
@brief	multiple BAM Pileup compression
@author	Yi Wang
@date	11/14/2010
*/
#include	<iostream>
#include	<gsl/gsl_cdf.h>
#include	"quality.cpp"
#include	<algorithm>
#include	<climits>
#include	<cstdlib>
#include	<fstream>
#include	<sstream>
#include	<string>
#include	<vector>
#include	<bam.h>
#include	<cmath>
using	namespace	std;
typedef	const	char	cchr;
typedef	float	fast;
typedef	double	real;
typedef	unsigned	uint;
const	fast	factor=1.0f/255.0f;

struct	bam_model{
	string	name;	//	bam file name
	bamFile	bamf;	//	bam file handle
	bam_iter_t	read;	//	bam read handle
	bam_plp_t	pbuf;	//	Pileup buffer
	const	bam_pileup1_t	*pile;	//	one Pileup
	int	cpos;	//	current position
	int	size;	//	read depth at cpos
	Quality	qua;	//	lr_model
};

class	Pileup{
private:
	vector<bam_model>	bam;
	string	reg,	chr;
	uint	bn;
	int	pos,	beg,	end;
	vector<char>	ref;
	uint8_t	code[256];
	char	name[16];
	gzFile	out;
	
	void	document();
	void	load_model(cchr	*F);	
	void	load_ref(cchr	*F);
	void	pileup(void);
	real	exact(uint	N[2][2]);
public:
	fast	square;
	void	options(int	ac,	char	**av);
	void	openr_bams(void);
	void	close_bams(void);
	void	work(void);

	static int	fetch_func(void *data, bam1_t *b);
};

int	Pileup::fetch_func(void *data,	bam1_t	*b){
	bam_model	*bm=(bam_model*)data;
	if(!bm->read)	return	-1;
	if(bam_iter_read(bm->bamf,	bm->read,	b)<0)	return	-1;
	bm->qua.adjust(b);
	return	1;
}

real	Pileup::exact(uint	N[2][2]){
	real	norm=lgamma(N[0][0]+N[0][1]+1)+lgamma(N[1][0]+N[1][1]+1)
	+lgamma(N[0][0]+N[1][0]+1)+lgamma(N[0][1]+N[1][1]+1)
	-lgamma(N[0][0]+N[0][1]+N[1][0]+N[1][1]+1);
	real	curr=exp(norm-lgamma(N[0][0]+1)-lgamma(N[0][1]+1)-lgamma(N[1][0]+1)-lgamma(N[1][1]+1)),	sum=0;
	uint	s=N[0][0]>N[1][1]?N[0][0]-N[1][1]:0;
	uint	e=N[0][1]<N[1][0]?N[0][0]+N[0][1]:N[0][0]+N[1][0];
	for(uint	i=s;	i<=e;	i++){
		real	p=exp(norm-lgamma(i+1)
		-lgamma(N[0][0]+N[0][1]-i+1)
		-lgamma(N[0][0]+N[1][0]-i+1)
		-lgamma(i+N[1][1]-N[0][0]+1));
		if(p<=curr)	sum+=p;
	}
	return	sum;
}

void	Pileup::pileup(void){
	if(pos>=(int)ref.size()||toupper(ref[pos])=='N')	return;
	uint8_t	refb=*(code+ref[pos]);
	vector<fast>	cnt;	cnt.resize(4*bn);
	real	ave[2]={0,0},	var[2]={0,0},	str[2][2]={{0,0},{0,0}};
	real	sum[4]={0,0,0,0};
	uint	dh[4]={0,0,0,0};

	for(uint	i=0;	i<bn;	i++)	if(bam[i].cpos==pos&&bam[i].size){
		bam_model	&bm=bam[i];
		fast	*c=&cnt[i*4];
		const	bam_pileup1_t	*p=bm.pile;
		for(int	k=0;	k<bm.size;	k++,	p++)	if(!p->is_del){
			uint8_t	call=*(code+bam1_seqi(bam1_seq(p->b),	p->qpos));
			fast	w=bam1_qual(p->b)[p->qpos]*factor;
			if(call==4)	continue;
			bool	type=(call==refb);
			c[call]+=w;	sum[call]+=w;
			if(bam1_strand(p->b))	str[type][1]+=w;	else	str[type][0]+=w;
			real	x=(p->qpos+0.5)/p->b->core.l_qseq;
			ave[type]+=w*x;	var[type]+=w*x*x;
		}
		for(uint	j=0;	j<4;	j++)	if(c[j]>square)	dh[j]++;
	}
	uint8_t	best=refb?0:1;
	for(uint	i=0;	i<4;	i++)	if(i!=refb&&sum[i]>sum[best])	best=i;

	if(dh[refb]<1||dh[best]<1)	return;
	
//	strand bias
	uint	istr[2][2];
	for(uint	i=0;	i<2;	i++)	for(uint	j=0;	j<2;	j++)	istr[i][j]=(uint)(str[i][j]+0.5);
	fast	pstr=exact(istr);
//	position bias
	real	rn=sum[refb],	an=sum[0]+sum[1]+sum[2]+sum[3]-rn;
	real	stdev=sqrt((var[0]-ave[0]*ave[0]/an+var[1]-ave[1]*ave[1]/rn)/(an+rn-2)*(1.0/an+1.0/rn));
	fast	ppos=gsl_cdf_tdist_P((ave[0]/an-ave[1]/rn)/stdev,	an+rn-2);
	ppos=2*fminf(ppos,	1-ppos);
	
	vector<uint16_t>	data;	data.resize(bn*2);
	for(uint	i=0;	i<bn;	i++){
		uint	r=(uint)(cnt[i*4+refb]*100+0.5),	a=(uint)(cnt[i*4+best]*100+0.5);
		data[i*2]=r<65536?r:65535;	data[i*2+1]=a<65536?a:65535;
	}
	uint32_t	temp=pos+1;
	gzwrite(out,	name,	16);
	gzwrite(out,	&temp,	4);
	gzwrite(out,	&refb,	1);	gzwrite(out,	&best,	1);
	gzwrite(out,	&pstr,	4);	gzwrite(out,	&ppos,	4);
	gzwrite(out,	&data[0],	4*bn);
}

void	Pileup::openr_bams(void){
	for(uint	i=0;	i<bn;	i++){
		if(!bam[i].qua.load(bam[i].name.c_str())){
			fprintf(stderr,	"no model avaliable for %s\n",	bam[i].name.c_str());
		}	
		bam[i].qua.refs=&ref[0];	bam[i].qua.refl=ref.size();

		bam[i].bamf=bam_open(bam[i].name.c_str(),	"r");
		if(bam[i].bamf==NULL){	fprintf(stderr,	"%s\tfailed to open %s\n",	__func__,	bam[i].name.c_str());	exit(0);	}
		bam_index_t	*idx=bam_index_load(bam[i].name.c_str());
		if(idx==0){	fprintf(stderr,	"%s\t%s\n",	__func__,	bam[i].name.c_str());	exit(0);	}
		bam_header_t	*header=bam_header_read(bam[i].bamf);
		int	tid;
		if(bam_parse_region(header,	reg.c_str(),	&tid,	&beg,	&end)<0)	
			fprintf(stderr,	"%s was not found in %s\n",	reg.c_str(),	bam[i].name.c_str());
		else{	
			bam[i].read=bam_iter_query(idx,	tid,	beg,	end);
			chr=header->target_name[tid];
		}
		bam_index_destroy(idx);	bam_header_destroy(header);
		bam[i].pbuf=bam_plp_init(fetch_func,	&bam[i]);
		bam[i].cpos=INT_MAX;
	}
	if(!chr.size()){	fprintf(stderr,	"%s\tregion was not found in all bams\n",	__func__);	exit(0);	}
}

void	Pileup::close_bams(void){
	for(uint	i=0;	i<bam.size();	i++){
		if(bam[i].read!=NULL)	bam_iter_destroy(bam[i].read);
		if(bam[i].pbuf!=NULL)	bam_plp_destroy(bam[i].pbuf);
		bam_close(bam[i].bamf);
	}
}

void	Pileup::load_ref(cchr	*F){
	ref.clear();
	FILE	*f=fopen(F,	"rt");
	if(f==NULL){	fprintf(stderr,	"can't open %s\n",	F);	exit(0);	}
	char	buff[1024];
	while(fgets(buff,	1024,	f)!=NULL)	if(buff[0]!='>'){
		for(char	*p=buff;	*p>33;	p++)	ref.push_back(*p);
	}
	fclose(f);
}

void	Pileup::work(void){
	bn=bam.size();
	for(uint	i=0;	i<256;	i++)	code[i]=4;
	code['A']=code['a']=code[1]=0;	code['C']=code['c']=code[2]=1;	
	code['G']=code['g']=code[4]=2;	code['T']=code['t']=code[8]=3;
	
	string	temp=reg+".gz";
	out=gzopen(temp.c_str(),	"wb");
	gzwrite(out,	&bn,	sizeof(uint));
	for(uint	i=0;	i<bn;	i++){
		uint	n=bam[i].name.size();
		gzwrite(out,	&n,	sizeof(uint));
		gzwrite(out,	bam[i].name.c_str(),	n);
	}
	
	openr_bams();	memset(name,	0,	16);	strcpy(name,	chr.c_str());
	pos=INT_MAX;
	for(;;){
		int	min=INT_MAX,	tid;
		for(uint	i=0;	i<bn;	i++){
			if(bam[i].cpos==pos)	bam[i].pile=bam_plp_auto(bam[i].pbuf,	&tid,	&bam[i].cpos,	&bam[i].size);
			if(bam[i].pile&&bam[i].cpos<min)	min=bam[i].cpos;
		}
		if(min==INT_MAX) break;
		pos=min;
		if(pos>=beg&&pos<end){
			if(!(pos%100000))	fprintf(stderr,	"%.1fM\r",	1e-6*pos);
			pileup();
		}	
	}
	close_bams();
	gzclose(out);
}

void	Pileup::document(void){
	fprintf(stderr,	"\nname	pileup 0.1");
	fprintf(stderr,	"\nfunc	multiple BAM pileup compression");
	fprintf(stderr,	"\nauthor	Yi Wang @ Fuli Yu' Group @ BCM");
	fprintf(stderr,	"\nusage	Pileup <region ref.fa square(1.9)> [bam1 bam2...]");
	fprintf(stderr,	"\n\n");
	exit(0);
}

void	Pileup::options(int	ac,	char	**av){
	if(ac<4)	document();
	reg=av[1];
	load_ref(av[2]);
	square=atof(av[3]);
	bam_model	temp;	temp.bamf=NULL;	temp.read=NULL;	temp.pile=NULL;
	for(int	i=4;	i<ac;	i++){
		temp.name=av[i];
		bam.push_back(temp);
	}
}

int	main(int	ac,	char	**av){
	Pileup	p;
	p.options(ac,	av);
	p.work();
	return	0;
}

